@@ -47,8 +47,6 @@
 #if DEBUG_RENDER
 #undef DBG
 #define DBG(x) ErrorF x
-#else
-#define NDEBUG 1
 #endif
 
 #define NO_COMPOSITE 0
@@ -344,22 +342,6 @@ static uint32_t gen7_get_card_format_for_depth(int depth)
 	}
 }
 
-static bool gen7_format_is_dst(uint32_t format)
-{
-	switch (format) {
-	case GEN7_SURFACEFORMAT_B8G8R8A8_UNORM:
-	case GEN7_SURFACEFORMAT_R8G8B8A8_UNORM:
-	case GEN7_SURFACEFORMAT_B10G10R10A2_UNORM:
-	case GEN7_SURFACEFORMAT_B5G6R5_UNORM:
-	case GEN7_SURFACEFORMAT_B5G5R5A1_UNORM:
-	case GEN7_SURFACEFORMAT_A8_UNORM:
-	case GEN7_SURFACEFORMAT_B4G4R4A4_UNORM:
-		return true;
-	default:
-		return false;
-	}
-}
-
 static uint32_t gen7_filter(uint32_t filter)
 {
 	switch (filter) {
@@ -788,7 +770,6 @@ gen7_emit_sf(struct sna *sna, Bool has_mask)
 		  1 << GEN7_SBE_URB_ENTRY_READ_LENGTH_SHIFT |
 		  1 << GEN7_SBE_URB_ENTRY_READ_OFFSET_SHIFT);
 	OUT_BATCH(0);
-	OUT_BATCH(0);
 	OUT_BATCH(0); /* dw4 */
 	OUT_BATCH(0);
 	OUT_BATCH(0);
@@ -1010,8 +991,7 @@ static void gen7_magic_ca_pass(struct sna *sna,
 		OUT_BATCH(MI_FLUSH | MI_INHIBIT_RENDER_CACHE_FLUSH);
 
 	OUT_BATCH(GEN7_3DPRIMITIVE | (7- 2));
-	OUT_BATCH(GEN7_3DPRIMITIVE_VERTEX_SEQUENTIAL |
-		  _3DPRIM_RECTLIST << GEN7_3DPRIMITIVE_TOPOLOGY_SHIFT);
+	OUT_BATCH(GEN7_3DPRIMITIVE_VERTEX_SEQUENTIAL | _3DPRIM_RECTLIST);
 	OUT_BATCH(sna->render.vertex_index - sna->render.vertex_start);
 	OUT_BATCH(sna->render.vertex_start);
 	OUT_BATCH(1);	/* single instance */
@@ -1210,26 +1190,16 @@ gen7_bind_bo(struct sna *sna,
 	if (is_dst) {
 		domains = I915_GEM_DOMAIN_RENDER << 16 |I915_GEM_DOMAIN_RENDER;
 		kgem_bo_mark_dirty(bo);
-	} else {
+	} else
 		domains = I915_GEM_DOMAIN_SAMPLER << 16;
-		is_dst = gen7_format_is_dst(format);
-	}
+
+	offset = kgem_bo_get_binding(bo, format);
+	if (offset)
+		return offset;
 
 	offset = sna->kgem.surface - sizeof(struct gen7_surface_state_padded) / sizeof(uint32_t);
 	offset *= sizeof(uint32_t);
 
-	if (is_dst) {
-		if (bo->dst_bound)
-			return bo->dst_bound;
-
-		bo->dst_bound = offset;
-	} else {
-		if (bo->src_bound)
-			return bo->src_bound;
-
-		bo->src_bound = offset;
-	}
-
 	sna->kgem.surface -=
 		sizeof(struct gen7_surface_state_padded) / sizeof(uint32_t);
 	ss = sna->kgem.batch + sna->kgem.surface;
@@ -1247,6 +1217,8 @@ gen7_bind_bo(struct sna *sna,
 	ss[6] = 0;
 	ss[7] = 0;
 
+	kgem_bo_set_binding(bo, format, offset);
+
 	DBG(("[%x] bind bo(handle=%d, addr=%d), format=%d, width=%d, height=%d, pitch=%d, tiling=%d -> %s\n",
 	     offset, bo->handle, ss[1],
 	     format, width, height, bo->pitch, bo->tiling,
@@ -1567,11 +1539,8 @@ static void gen7_emit_primitive(struct sna *sna)
 		return;
 	}
 
-	OUT_BATCH(GEN7_3DPRIMITIVE |
-		  GEN7_3DPRIMITIVE_VERTEX_SEQUENTIAL |
-		  _3DPRIM_RECTLIST << GEN7_3DPRIMITIVE_TOPOLOGY_SHIFT |
-		  0 << 9 |
-		  4);
+	OUT_BATCH(GEN7_3DPRIMITIVE | (7- 2));
+	OUT_BATCH(GEN7_3DPRIMITIVE_VERTEX_SEQUENTIAL | _3DPRIM_RECTLIST);
 	sna->render_state.gen7.vertex_offset = sna->kgem.nbatch;
 	OUT_BATCH(0);	/* vertex count, to be filled in later */
 	OUT_BATCH(sna->render.vertex_index);
@@ -2208,8 +2177,8 @@ gen7_composite_set_target(struct sna_composite_op *op, PicturePtr dst)
 			return FALSE;
 
 		op->dst.bo = priv->gpu_bo;
-		if (!priv->gpu_only &&
-		    !sna_damage_is_all(&priv->gpu_damage, op->dst.width, op->dst.height))
+		if (!sna_damage_is_all(&priv->gpu_damage,
+				       op->dst.width, op->dst.height))
 			op->damage = &priv->gpu_damage;
 	}
 
@@ -2254,6 +2223,8 @@ gen7_render_composite(struct sna *sna,
 		      int16_t width, int16_t height,
 		      struct sna_composite_op *tmp)
 {
+	if (op >= ARRAY_SIZE(gen7_blend_op))
+		return FALSE;
 
 #if NO_COMPOSITE
 	if (mask)
@@ -2278,9 +2249,6 @@ gen7_render_composite(struct sna *sna,
 			      width, height, tmp))
 		return TRUE;
 
-	if (op >= ARRAY_SIZE(gen7_blend_op))
-		return FALSE;
-
 	if (need_tiling(sna, width, height))
 		return sna_tiling_composite(op, src, mask, dst,
 					    src_x, src_y,
@@ -2292,6 +2260,7 @@ gen7_render_composite(struct sna *sna,
 	tmp->op = op;
 	if (!gen7_composite_set_target(tmp, dst))
 		return FALSE;
+	sna_render_reduce_damage(tmp, dst_x, dst_y, width, height);
 
 	if (tmp->dst.width > GEN7_MAX_SIZE || tmp->dst.height > GEN7_MAX_SIZE) {
 		if (!sna_render_composite_redirect(sna, tmp,
