@@ -42,24 +42,12 @@
 #if DEBUG_VIDEO_TEXTURED
 #undef DBG
 #define DBG(x) ErrorF x
-#else
-#define NDEBUG 1
 #endif
 
 #define MAKE_ATOM(a) MakeAtom(a, sizeof(a) - 1, TRUE)
 
 static Atom xvBrightness, xvContrast, xvSyncToVblank;
 
-/* client libraries expect an encoding */
-static const XF86VideoEncodingRec DummyEncoding[1] = {
-	{
-		0,
-		"XV_IMAGE",
-		8192, 8192,
-		{1, 1}
-	}
-};
-
 #define NUM_FORMATS 3
 
 static XF86VideoFormatRec Formats[NUM_FORMATS] = {
@@ -67,11 +55,11 @@ static XF86VideoFormatRec Formats[NUM_FORMATS] = {
 };
 
 //#define NUM_TEXTURED_ATTRIBUTES 3
-#define NUM_TEXTURED_ATTRIBUTES 0
+#define NUM_TEXTURED_ATTRIBUTES 1
 static XF86AttributeRec TexturedAttributes[] = {
+	{XvSettable | XvGettable, -1, 1, "XV_SYNC_TO_VBLANK"},
 	{XvSettable | XvGettable, -128, 127, "XV_BRIGHTNESS"},
 	{XvSettable | XvGettable, 0, 255, "XV_CONTRAST"},
-	{XvSettable | XvGettable, -1, 1, "XV_SYNC_TO_VBLANK"},
 };
 
 #ifdef SNA_XVMC
@@ -239,22 +227,20 @@ sna_video_textured_put_image(ScrnInfoPtr scrn,
 	PixmapPtr pixmap = get_drawable_pixmap(drawable);
 	BoxRec dstBox;
 	xf86CrtcPtr crtc;
-	int top, left, npixels, nlines;
 	Bool flush = false;
 
 	if (!sna_pixmap(pixmap))
 		return BadAlloc;
 
-	if (!sna_video_clip_helper(scrn, video, &crtc, &dstBox,
+	sna_video_frame_init(sna, video, id, width, height, &frame);
+
+	if (!sna_video_clip_helper(scrn, video, &frame,
+				   &crtc, &dstBox,
 				   src_x, src_y, drw_x, drw_y,
 				   src_w, src_h, drw_w, drw_h,
-				   id,
-				   &top, &left, &npixels, &nlines,
-				   clip, width, height))
+				   clip))
 		return Success;
 
-	sna_video_frame_init(sna, video, id, width, height, &frame);
-
 	if (xvmc_passthrough(id)) {
 		if (IS_I915G(sna) || IS_I915GM(sna)) {
 			/* XXX: i915 is not support and needs some
@@ -266,10 +252,14 @@ sna_video_textured_put_image(ScrnInfoPtr scrn,
 		if (frame.bo == NULL)
 			return BadAlloc;
 	} else {
-		if (!sna_video_copy_data(sna, video, &frame,
-					 top, left, npixels, nlines,
-					 buf))
+		frame.bo = kgem_create_linear(&sna->kgem, frame.size);
+		if (frame.bo == NULL)
 			return BadAlloc;
+
+		if (!sna_video_copy_data(sna, video, &frame, buf)) {
+			kgem_bo_destroy(&sna->kgem, frame.bo);
+			return BadAlloc;
+		}
 	}
 
 	if (crtc && video->SyncToVblank != 0)
@@ -281,7 +271,7 @@ sna_video_textured_put_image(ScrnInfoPtr scrn,
 			  drw_w, drw_h,
 			  pixmap);
 
-	sna_video_frame_fini(sna, video, &frame);
+	kgem_bo_destroy(&sna->kgem, frame.bo);
 
 	DamageDamageRegion(drawable, clip);
 
@@ -408,8 +398,13 @@ XF86VideoAdaptorPtr sna_video_textured_setup(struct sna *sna,
 	adaptor->flags = 0;
 	adaptor->name = "Intel(R) Textured Video";
 	adaptor->nEncodings = 1;
-	adaptor->pEncodings = xnfalloc(sizeof(DummyEncoding));
-	memcpy(adaptor->pEncodings, DummyEncoding, sizeof(DummyEncoding));
+	adaptor->pEncodings = xnfalloc(sizeof(XF86VideoEncodingRec));
+	adaptor->pEncodings[0].id = 0;
+	adaptor->pEncodings[0].name = "XV_IMAGE";
+	adaptor->pEncodings[0].width = sna->render.max_3d_size;
+	adaptor->pEncodings[0].height = sna->render.max_3d_size;
+	adaptor->pEncodings[0].rate.numerator = 1;
+	adaptor->pEncodings[0].rate.denominator = 1;
 	adaptor->nFormats = NUM_FORMATS;
 	adaptor->pFormats = Formats;
 	adaptor->nPorts = nports;
