@@ -111,6 +111,9 @@ static void uxa_unrealize_glyph_caches(ScreenPtr pScreen)
 	uxa_screen_t *uxa_screen = uxa_get_screen(pScreen);
 	int i;
 
+	if (!uxa_screen->glyph_cache_initialized)
+		return;
+
 	for (i = 0; i < UXA_NUM_GLYPH_CACHE_FORMATS; i++) {
 		uxa_glyph_cache_t *cache = &uxa_screen->glyphCaches[i];
 
@@ -120,6 +123,7 @@ static void uxa_unrealize_glyph_caches(ScreenPtr pScreen)
 		if (cache->glyphs)
 			free(cache->glyphs);
 	}
+	uxa_screen->glyph_cache_initialized = FALSE;
 }
 
 void uxa_glyphs_fini(ScreenPtr pScreen)
@@ -145,6 +149,10 @@ static Bool uxa_realize_glyph_caches(ScreenPtr pScreen)
 	};
 	int i;
 
+	if (uxa_screen->glyph_cache_initialized)
+		return TRUE;
+
+	uxa_screen->glyph_cache_initialized = TRUE;
 	memset(uxa_screen->glyphCaches, 0, sizeof(uxa_screen->glyphCaches));
 
 	for (i = 0; i < sizeof(formats)/sizeof(formats[0]); i++) {
@@ -214,17 +222,7 @@ Bool uxa_glyphs_init(ScreenPtr pScreen)
 	if (uxa_get_screen(pScreen)->force_fallback)
 		return TRUE;
 
-	/* We are trying to initialise per screen resources prior to the
-	 * complete initialisation of the screen. So ensure the components
-	 * that we depend upon are initialsed prior to our use.
-	 */
-	if (!CreateScratchPixmapsForScreen(pScreen->myNum))
-		return FALSE;
-
-	if (!uxa_realize_glyph_caches(pScreen))
-		return FALSE;
-
-	return TRUE;
+	return uxa_realize_glyph_caches(pScreen);
 }
 
 /* The most efficient thing to way to upload the glyph to the screen
@@ -939,6 +937,11 @@ uxa_glyphs_via_mask(CARD8 op,
 	if (!pixmap)
 		return 1;
 
+	if (!uxa_pixmap_is_offscreen(pixmap)) {
+		screen->DestroyPixmap(pixmap);
+		return -1;
+	}
+
 	uxa_clear_pixmap(screen, uxa_screen, pixmap);
 
 	component_alpha = NeedsComponent(maskFormat->format);
@@ -1030,6 +1033,19 @@ next_glyph:
 	return 0;
 }
 
+static Bool
+is_solid(PicturePtr picture)
+{
+	if (picture->pSourcePict) {
+		SourcePict *source = picture->pSourcePict;
+		return source->type == SourcePictTypeSolidFill;
+	} else {
+		return (picture->repeat &&
+			picture->pDrawable->width  == 1 &&
+			picture->pDrawable->height == 1);
+	}
+}
+
 void
 uxa_glyphs(CARD8 op,
 	   PicturePtr pSrc,
@@ -1050,7 +1066,9 @@ uxa_glyphs(CARD8 op,
 	    uxa_screen->swappedOut ||
 	    uxa_screen->force_fallback ||
 	    !uxa_drawable_is_offscreen(pDst->pDrawable) ||
-	    pDst->alphaMap || pSrc->alphaMap) {
+	    pDst->alphaMap || pSrc->alphaMap ||
+	    /* XXX we fail to handle (rare) non-solid sources correctly. */
+	    !is_solid(pSrc)) {
 fallback:
 	    uxa_check_glyphs(op, pSrc, pDst, maskFormat, xSrc, ySrc, nlist, list, glyphs);
 	    return;
@@ -1131,6 +1149,11 @@ fallback:
 		if (!pixmap)
 			return;
 
+		if (!uxa_pixmap_is_offscreen(pixmap)) {
+			screen->DestroyPixmap(pixmap);
+			goto fallback;
+		}
+
 		gc = GetScratchGC(depth, screen);
 		if (!gc) {
 			screen->DestroyPixmap(pixmap);
